#include <sys/stat.h>
#include <sys/wait.h>
#include "np.h"

#define MAXFD      64
#define TRANSFER_BUFFER_SIZE     10000
#define LISTENQ	1024

/*
 * argv[1] = source port;
 * argv[2] = destination server;
 * argv[3] = desintation port;
 */

void *Proxy (void *ptr);
int proxy (int fd, char *szHost, char *szPort);
void *TransferThread (void *data);
void sig_chld (int signo);
int daemon_init (const char *pname, int facility);

typedef struct tag_transfer
  {
    int fdSource;
    int fdDest;
  }
Transfer;

struct proxy_t
  {
    char *ip;
    char *port;
    int fd;
  };

extern int daemon_proc;

/*
 * int main()
 */
int 
main (int argc, char *argv[])
{
  int listenfd, connfd;
  socklen_t clilen;
  socklen_t addrlen;
  struct sockaddr_in cliaddr;
  struct proxy_t *ptr;
  pthread_t ptConnection;

  if (argc < 4)
    {
      printf ("Usage: proxy [listen port][dest server][dest port]\n");
      exit (0);
    }

  ptr = (struct proxy_t *) malloc (sizeof (struct proxy_t));

  ptr->ip = argv[2];
  ptr->port = argv[3];

  /* Run proxy as UNIX daemon  -SP 12/23/98 */
  daemon_init (argv[0], 6);

  listenfd = tcp_listen (NULL, argv[1], &addrlen);

  for (;;)
    {
      clilen = sizeof (cliaddr);
      connfd = accept (listenfd, (SA *) & cliaddr, &clilen);
      if (connfd == -1)
	{
	  continue;
	}

      ptr->fd = connfd;

      if (pthread_create (&ptConnection,
			  NULL,
			  Proxy,
			  (void *) ptr) != 0)
	{
	  perror ("Thread creation problems");
	  exit (1);
	}
    }
}

/*
 * void *Proxy()
 *
 * Extract function parameters from structure pointer.
 */
void *
Proxy (void *ptr)
{
  int connfd;
  struct proxy_t *pt;

  pt = (struct proxy_t *) ptr;
  connfd = pt->fd;
  proxy (connfd, pt->ip, pt->port);
  return NULL;
}

/*
 * void proxy()
 *
 * Description:
 * Proxy function.  Proxies all data from file descriptor
 * to the specified host.  The file descriptor is assumed
 * to be socket opened with any socket tools.
 */
int 
proxy (int fd, char *szHost, char *szPort)
{
  int fdOut;
  pthread_t ptWrite;
  Transfer tRead, tWrite;

  fdOut = tcp_connect (szHost, szPort);

  /* Split to process incomming traffic and outgoing traffic... */
  /* One thread for each train of traffic.                      */

  /* Create a thread for the Writing to our Out server.         */
  tWrite.fdSource = fd;
  tWrite.fdDest = fdOut;

  if (pthread_create (&ptWrite,
		      NULL,
		      TransferThread,
		      &tWrite) != 0)
    {
      perror ("thread creation problems");
      exit (1);
    }

  tRead.fdSource = fdOut;
  tRead.fdDest = fd;
  TransferThread (&tRead);

  pthread_cancel (ptWrite);	/* Get rid of thread after proxy */
  return 0;
}

/*
 * void *TransferThead()
 * 
 * Description:
 * This function actually performs the data transfer
 * between two socket descriptors. It should be passed 
 * the Transfer structure.  fdSource and fdDest will the 
 * be source and destination file descriptors.  This 
 * function just loops reading data from fdSource and 
 * writing it to fdDest.
 */
void *
TransferThread (void *data)
{
  Transfer *tControl;
  char szBuffer[TRANSFER_BUFFER_SIZE];
  int nReadAmt, nWriteAmt;

  tControl = data;

  for (;;)
    {
      nReadAmt = read (tControl->fdSource, &szBuffer, TRANSFER_BUFFER_SIZE);

      if (nReadAmt <= 0)
	{
	  if (close (tControl->fdDest) == -1)
	    perror ("Close problems");
	  if (close (tControl->fdSource) == -1)
	    perror ("Close problems");
	  return NULL;
	}

      while (nReadAmt > 0)
	{
	  nWriteAmt = write (tControl->fdDest, &szBuffer, nReadAmt);
	  if (nWriteAmt <= 0)
	    return NULL;

	  nReadAmt -= nReadAmt;
	}
    }
  return data;
}

/*
 * int daemon_init()
 *
 * Causes proxy to run as a UNIX daemon
 */
int 
daemon_init (const char *pname, int facility)
{
  int i;
  pid_t pid;

  if ((pid = fork ()) != 0)
    exit (0);

  setsid ();

  signal (SIGHUP, SIG_IGN);

  if ((pid = fork ()) != 0)
    exit (0);

  chdir ("/");

  umask (0);

  for (i = 0; i < MAXFD; i++)
    close (i);

  openlog (pname, LOG_PID, facility);

  return 0;
}
